#include <stdlib.h>
#include <stdio.h>
#include "BinHeap.h"
#include "AppleSolver.h"

AppleVectorT *createApples(int size)
{
    AppleVectorT *v = (AppleVectorT *)malloc(sizeof(AppleVectorT));
    if (!v)
    {
        abort();
    }
    v->apples = (AppleT **)malloc(size * sizeof(AppleT *));
    if (!v->apples)
    {
        abort();
    }
    v->size = size;
    return v;
}

AppleT *createApple(int h, int w)
{
    AppleT *a = (AppleT *)malloc(sizeof(AppleT));
    if (!a)
    {
        abort();
    }
    a->h = h;
    a->w = w;
    return a;
}

AppleVectorT *cloneAppleVector(AppleVectorT *v)
{
    AppleVectorT *vClone = createApples(v->size);
    if (!vClone)
    {
        abort();
    }
    for (int i=0; i < v->size; i++)
        vClone->apples[i] = createApple(v->apples[i]->h, v->apples[i]->w);
    return vClone;
}

static int sum(AppleVectorT *apples, int H)
{
    int sum = 0;
    AppleT *
    *ap = apples->apples;
    for (int i=0; i < apples->size; i++)
        if (ap[i]->h < H)
            sum += ap[i]->w;
    return sum;
}

static int cmpApples(const void *a1, const void *a2)
{
    int w1 = (*((AppleT **)a1))->w;
    int w2 = (*((AppleT **)a2))->w;
    // reversed for decreasing order
    return w1 - w2;
}

static int cmpAppleWeights(const void *a1, const void *a2)
{
    return *(int *)a1-*(int *)a2;
}

void printApples(AppleVectorT *apples)
{
    printf("AppleVector: %d [", apples->size);
    for (int i=0; i < apples->size; i++)
    {
        if (i) printf(" ");
        printf("(%d, %d)", apples->apples[i]->w, apples->apples[i]->h);
    }
    printf("]\n");
}
void binHeapPrint(BinHeapT *heap)
{
    printf("WeightsHeap: %d [", binHeapGetCount(heap));
    for (int i=0; i < heap->count; i++)
    {
        if (i) putchar(' ');
        printf("%d", *(int *)heap->values[i]);
    }
    printf("]\n");

}

int solve(AppleVectorT *apples, int H, int U)
{
    if (U == 0)
    {
        return sum(apples, H);
    }
    // sort apples in ascending order
    qsort(apples->apples, apples->size, sizeof(AppleT *), cmpApples);

    printf("Sort:\n");
    printApples(apples);

    // no apple picked, yet
    int pickedWeight = 0;

    // create a max PQ
    BinHeapT *availableWeights = binHeapCreate(BIN_HEAP_TYPE_MAX, cmpAppleWeights);

    int offset = U - H % U;
    if (offset == U) offset = 0;
    int top = offset - U;

    while ((apples->size || binHeapGetCount(availableWeights)) && top <= H)
    {
        if (binHeapGetCount(availableWeights))
        {   // can take an apple
            pickedWeight += *(int *)binHeapPop(availableWeights);
            top += U; // branch lifts
        }
        else
        {
            int val = (apples->apples[apples->size-1]->h - top) / U;
            if ( val < 1 ) val = 1;
            top += U * val;
        }
//        printf("If:\n");
//        binHeapPrint(availableWeights);
//        printf("top=%d\n", top);
        while (apples->size && apples->apples[apples->size-1]->h <= top)
        {   // still have apples and are reachable
            binHeapInsert(availableWeights, &apples->apples[apples->size-1]->w);
            --apples->size;
        }
//        printf("While:\n");
//        printApples(apples);
//        binHeapPrint(availableWeights);
    }
    return pickedWeight;
}
